/*
 * Tormenta 2  Quad-E1/T1 PCI Driver
 *
 * Copyright (C) 2005, Martin Pycko
 * Written by Martin Pycko <mpycko@pbxhardware.com>
 *
 * Buy the T1/E1 hardware that works with this driver at
 * http://pbxhardware.com
 *
 * We offer dual E1, dual T1, quad E1, quad T1 and quad T1/E1 hardware
 *
 * NOTE:
 * This driver cannot be redistributed unless all the header information
 * gets preserved intact!
 *
 * Derived from work by:
 * Written by Mark Spencer <markster@digium.com>
 * Based on previous works, designs, and archetectures conceived and
 * written by Jim Dixon <jim@lambdatel.com>.
 *
 * Copyright (C) 2001 Jim Dixon / Zapata Telephony.
 * Copyright (C) 2001-2008, Digium, Inc.
 *
 * Adapted driver for E1/T1 operation by Luis Vazquez <luis@ipcontact.com.uy>
 * DAHDI version by Daniel Ferrer <daniel@ipcontact.com.uy>
 *
 */

/*
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2 as published by the
 * Free Software Foundation. See the LICENSE file included with
 * this program for more details.
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/moduleparam.h>

#include <dahdi/kernel.h>
#define NEED_PCI_IDS
#include "tor3-hw.h"
#include "tor2fw.h"

/*
 * Tasklets provide better system interactive response at the cost of the
 * possibility of losing a frame of data at very infrequent intervals.  If
 * you are more concerned with the performance of your machine, enable the
 * tasklets.  If you are strict about absolutely no drops, then do not enable
 * tasklets.
 */

/* #define ENABLE_TASKLETS */

/* this stuff needs to work for 64 bit systems, however using the macro causes
   it to take twice as long */
/* #define FIXTHISFOR64 */  /* as of now, un-comment for 32 bit only system */

#define SPANS_PER_CARD  4
#define MAX_SPANS       16

#define FLAG_STARTED (1 << 0)

#define	TYPE_T1	1		/* is a T1 card */
#define	TYPE_E1	2		/* is an E1 card */

struct tor3_chan {
	/* Private pointer for channel.  We want to know our
	   channel and span */
	struct tor3 *tor;
	int span;	/* Index from 0 */
};

struct tor3_span {
	/* Private pointer for span.  We want to know our
	   span number and pointer to the tor device */
	struct tor3 *tor;
	int span;	/* Index from 0 */
	unsigned int *revdatxlt;	/* pointer to datxlt structure */
	int cardtype;		/* card type, T1 or E1 */
};

struct tor3 {
	/* This structure exists one per card */
	struct pci_dev *pci;		/* Pointer to PCI device */
	int num;			/* Which card we are */
	int syncsrc;			/* active sync source */
	int syncs[SPANS_PER_CARD];	/* sync sources */
	int psyncs[SPANS_PER_CARD];	/* span-relative sync sources */
	int alarmtimer[SPANS_PER_CARD];	/* Alarm timer */
	char *type;			/* Type of tormenta 2 card */
	int irq;			/* IRQ used by device */
	int order;			/* Order */
	int flags;			/* Device flags */
	int syncpos[SPANS_PER_CARD];	/* span-relative sync sources */
	int master;			/* Are we master */
	unsigned long plx_region;	/* phy addr of PCI9030 registers */
	unsigned long plx_len;		/* length of PLX window */
	__iomem volatile unsigned short *plx;	/* Virtual representation of local space */
	unsigned long xilinx32_region;	/* 32 bit Region allocated to Xilinx */
	unsigned long xilinx32_len;	/* Length of 32 bit Xilinx region */
	__iomem volatile unsigned int *mem32;	/* Virtual representation of 32 bit Xilinx memory area */
	unsigned long xilinx8_region;	/* 8 bit Region allocated to Xilinx */
	unsigned long xilinx8_len;	/* Length of 8 bit Xilinx region */
	__iomem volatile unsigned char *mem8;	/* Virtual representation of 8 bit Xilinx memory area */
	struct dahdi_span spans[SPANS_PER_CARD];		/* Spans */
	struct tor3_span tspans[SPANS_PER_CARD];	/* Span data */
	struct dahdi_chan **chans[SPANS_PER_CARD]; /* Pointers to card channels */
	struct tor3_chan tchans[32 * SPANS_PER_CARD];	/* Channel user data */
	unsigned char txsigs[SPANS_PER_CARD][16];	/* Copy of tx sig registers */
	int loopupcnt[SPANS_PER_CARD];	/* loop up code counter */
	int loopdowncnt[SPANS_PER_CARD];/* loop down code counter */
	int spansstarted;		/* number of spans started */
	spinlock_t lock;		/* lock context */
	unsigned char leds;		/* copy of LED register */
	unsigned char ec_chunk1[SPANS_PER_CARD][32][DAHDI_CHUNKSIZE]; /* first EC chunk buffer */
	unsigned char ec_chunk2[SPANS_PER_CARD][32][DAHDI_CHUNKSIZE]; /* second EC chunk buffer */
#ifdef ENABLE_TASKLETS
	int taskletrun;
	int taskletsched;
	int taskletpending;
	int taskletexec;
	int txerrors;
	struct tasklet_struct tor3_tlet;
#endif
	unsigned int *datxlt;	/* pointer to datxlt structure */
	int e1div;		/* should we add e1div */
	int lastmode;
	unsigned int passno;	/* number of interrupt passes */
};

#define t1out(tor,span,reg,val) tor->mem8[((span - 1) * 0x100) + reg] = val
#define t1in(tor,span,reg) tor->mem8[((span - 1) * 0x100) + reg]

#ifdef ENABLE_TASKLETS
static void tor3_tasklet(unsigned long data);
#endif

#define	GPIOC (PLX_LOC_GPIOC >> 1) /* word-oriented address for PLX GPIOC reg. (32 bit reg.) */
#define	LAS2BRD (0x30 >> 1)
#define	LAS3BRD (0x34 >> 1)
#define	INTCSR (0x4c >> 1)  /* word-oriented address for PLX INTCSR reg. */
#define	PLX_INTENA 0x43 /* enable, hi-going, level trigger */

#define	SYNCREG	0x400
#define	CTLREG	0x401
#define	LEDREG	0x402
#define	STATREG	0x400
#define SWREG	0x401
#define	CTLREG1	0x404

#define	INTENA	(1 + ((loopback & 3) << 5))
#define	OUTBIT	(2 + ((loopback & 3) << 5))
#define	E1DIV	0x10
#define	INTACK	(0x80 + ((loopback & 3) << 5))
#define	INTACTIVE 2
#define MASTER (1 << 3)

/* un-define this if you dont want NON-REV A hardware support */
/* #define	NONREVA 1 */

#define	SYNCSELF 0
#define	SYNC1	1
#define	SYNC2	2
#define	SYNC3	3
#define	SYNC4	4
#define SYNCEXTERN 5

#define	LEDRED	2
#define	LEDGREEN 1

#define MAX_TOR_CARDS 64

static struct tor3 *cards[MAX_TOR_CARDS];

/* signalling bits */
#define	TOR_ABIT 8
#define	TOR_BBIT 4

static int debug;
static int japan;
#ifdef DEFAULT_TYPE_E1
static int mode = 0;
#else
static int mode = 15;
#endif
static int loopback;
static int highestorder;
static int timingcable;

static void set_clear_t1(struct tor3 *tor, int s);
static void set_clear_e1(struct tor3 *tor, int s);
static int tor3_startup(struct dahdi_span *span);
static int tor3_shutdown(struct dahdi_span *span);
static int tor3_rbsbits(struct dahdi_chan *chan, int bits);
static int tor3_maint(struct dahdi_span *span, int cmd);
static int tor3_ioctl(struct dahdi_chan *chan, unsigned int cmd, unsigned long data);
DAHDI_IRQ_HANDLER(tor3_intr);

/* translations of data channels for 24 channels in a 32 bit PCM highway */
static unsigned datxlt_t1[] = { 
    1 ,2 ,3 ,5 ,6 ,7 ,9 ,10,11,13,14,15,17,18,19,21,22,23,25,26,27,29,30,31 };
unsigned revdatxlt_t1[] = {
    1, 2, 3, 4, 4, 5, 6, 7, 7, 8,9,10,10,11,12,13,13,14,15,16,16,17,18,19,
	19,20,21,22,22,23,24 }; 

/* translations of data channels for 30/31 channels in a 32 bit PCM highway */
static unsigned datxlt_e1[] = { 
    1 ,2 ,3 ,4 ,5 ,6 ,7 ,8 ,9 ,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,
	25,26,27,28,29,30,31 };

static inline int DOSYNCREG(struct tor3 *tor, int mode)
{
	if (tor->lastmode != mode) {
		tor->lastmode = mode;
		if (mode > 0 && mode <= 4) {
			if (tor->tspans[mode - 1].cardtype == TYPE_T1) {
				tor->e1div = 0;
			} else {
				tor->e1div = E1DIV;
			}
		} else {
			tor->e1div = E1DIV;
		}
		tor->mem8[SYNCREG] = mode;
	}
	return 0;
}

static int tor3_spanconfig(struct dahdi_span *span, struct dahdi_lineconfig *lc)
{
	int i;
	struct tor3_span *p = span->pvt;

	if (debug)
		printk(KERN_INFO "Tor3: Configuring span %d\n", span->spanno);

	if ((lc->sync < 0) || (lc->sync >= SPANS_PER_CARD)) {
		printk(KERN_WARNING "%s %d: invalid span timing value %d.\n",
				THIS_MODULE->name, span->spanno, lc->sync);
		return -EINVAL;
	}
	/* XXX We assume lineconfig is okay and shouldn't XXX */	
	span->lineconfig = lc->lineconfig;
	span->txlevel = lc->lbo;
	span->rxlevel = 0;
	span->syncsrc = p->tor->syncsrc;
	
	/* remove this span number from the current sync sources, if there */
	for (i = 0; i < SPANS_PER_CARD; i++) {
		if (p->tor->syncs[i] == span->spanno) {
			p->tor->syncs[i] = 0;
			p->tor->psyncs[i] = 0;
		}
	}
	p->tor->syncpos[p->span] = lc->sync;
	/* if a sync src, put it in the proper place */
	if (lc->sync) {
		p->tor->syncs[lc->sync - 1] = span->spanno;
		p->tor->psyncs[lc->sync - 1] = p->span + 1;
	}
	/* If we're already running, then go ahead and apply the changes */
	if (span->flags & DAHDI_FLAG_RUNNING)
		return tor3_startup(span);

	return 0;
}

static int tor3_chanconfig(struct dahdi_chan *chan, int sigtype)
{
	int alreadyrunning;
	unsigned long flags;
	struct tor3_chan *p = chan->pvt;

	alreadyrunning = chan->span->flags & DAHDI_FLAG_RUNNING;
	if (debug) {
		if (alreadyrunning)
			printk(KERN_INFO "Tor3: Reconfigured channel %d (%s) sigtype %d\n", chan->channo, chan->name, sigtype);
		else
			printk(KERN_INFO "Tor3: Configured channel %d (%s) sigtype %d\n", chan->channo, chan->name, sigtype);
	}		
	spin_lock_irqsave(&p->tor->lock, flags);	
	if (alreadyrunning) {
		if (p->tor->tspans[p->span].cardtype == TYPE_E1)
			set_clear_e1(p->tor, p->span);
		else
			set_clear_t1(p->tor, p->span);
	}
	spin_unlock_irqrestore(&p->tor->lock, flags);	
	return 0;
}

static int tor3_open(struct dahdi_chan *chan)
{
	return 0;
}

static int tor3_close(struct dahdi_chan *chan)
{
	return 0;
}

static void init_spans(struct tor3 *tor)
{
	int x, y, c;
	/* TODO: a debug printk macro */
	for (x = 0; x < SPANS_PER_CARD; x++) {
		sprintf(tor->spans[x].name, "Tor3/%d/%d", tor->num, x + 1);
		snprintf(tor->spans[x].desc, sizeof(tor->spans[x].desc) - 1,
			 "Tormenta 2 (PCI) Quad %s Card %d Span %d",
			 (tor->tspans[x].cardtype == TYPE_T1)  ?  "T1"  :  "E1", tor->num, x + 1);
		printk(KERN_INFO "%s\n", tor->spans[x].desc);
		snprintf(tor->spans[x].location, sizeof(tor->spans[x].location) - 1,
			 "PCI Bus %02d Slot %02d", tor->pci->bus->number, PCI_SLOT(tor->pci->devfn) + 1);
		tor->spans[x].spanconfig = tor3_spanconfig;
		tor->spans[x].chanconfig = tor3_chanconfig;
		tor->spans[x].startup = tor3_startup;
		tor->spans[x].shutdown = tor3_shutdown;
		tor->spans[x].rbsbits = tor3_rbsbits;
		tor->spans[x].maint = tor3_maint;
		tor->spans[x].open = tor3_open;
		tor->spans[x].close  = tor3_close;
		if (tor->tspans[x].cardtype == TYPE_T1) {
			tor->spans[x].channels = 24;
			tor->spans[x].deflaw = DAHDI_LAW_MULAW;
			tor->tspans[x].revdatxlt = revdatxlt_t1;
		} else {
			tor->spans[x].channels = 31;
			tor->spans[x].deflaw = DAHDI_LAW_ALAW;
			tor->tspans[x].revdatxlt = datxlt_e1;
		}
		tor->spans[x].chans = tor->chans[x];
		tor->spans[x].flags = DAHDI_FLAG_RBS;
		tor->spans[x].linecompat = DAHDI_CONFIG_AMI | DAHDI_CONFIG_B8ZS | DAHDI_CONFIG_D4 | DAHDI_CONFIG_ESF | DAHDI_CONFIG_CCS | DAHDI_CONFIG_HDB3 | DAHDI_CONFIG_CRC4;
		tor->spans[x].ioctl = tor3_ioctl;
		tor->spans[x].pvt = &tor->tspans[x];
		tor->tspans[x].tor = tor;
		tor->tspans[x].span = x;
		init_waitqueue_head(&tor->spans[x].maintq);
		for (y=0;y<tor->spans[x].channels;y++) {
			struct dahdi_chan *mychans = tor->chans[x][y];
			sprintf(mychans->name, "Tor3/%d/%d/%d", tor->num, x + 1, y + 1);
			mychans->sigcap = DAHDI_SIG_EM | DAHDI_SIG_CLEAR | DAHDI_SIG_FXSLS | DAHDI_SIG_FXSGS | DAHDI_SIG_FXSKS |
 									 DAHDI_SIG_FXOLS | DAHDI_SIG_FXOGS | DAHDI_SIG_FXOKS | DAHDI_SIG_CAS | DAHDI_SIG_SF | DAHDI_SIG_EM_E1 | DAHDI_SIG_DACS_RBS;
			c = (x * tor->spans[x].channels) + y;
			mychans->pvt = &tor->tchans[c];
			mychans->chanpos = y + 1;
			tor->tchans[c].span = x;
			tor->tchans[c].tor = tor;
		}
	}
}

static int __devinit tor3_launch(struct tor3 *tor)
{
	if (test_bit(DAHDI_FLAGBIT_REGISTERED, &tor->spans[0].flags))
		return 0;
	printk(KERN_INFO "Tor3: Launching card: %d\n", tor->order);
	if (dahdi_register(&tor->spans[0], 0)) {
		printk(KERN_ERR "Unable to register span %s\n", tor->spans[0].name);
		return -1;
	}
	if (dahdi_register(&tor->spans[1], 0)) {
		printk(KERN_ERR "Unable to register span %s\n", tor->spans[1].name);
		dahdi_unregister(&tor->spans[0]);
		return -1;
	}
	if (dahdi_register(&tor->spans[2], 0)) {
		printk(KERN_ERR "Unable to register span %s\n", tor->spans[2].name);
		dahdi_unregister(&tor->spans[0]);
		dahdi_unregister(&tor->spans[1]);
		return -1;
	}
	if (dahdi_register(&tor->spans[3], 0)) {
		printk(KERN_ERR "Unable to register span %s\n", tor->spans[3].name);
		dahdi_unregister(&tor->spans[0]);
		dahdi_unregister(&tor->spans[1]);
		dahdi_unregister(&tor->spans[2]);
		return -1;
	}
	tor->plx[INTCSR] = cpu_to_le16(PLX_INTENA); /* enable PLX interrupt */
#ifdef ENABLE_TASKLETS
	tasklet_init(&tor->tor3_tlet, tor3_tasklet, (unsigned long)tor);
#endif
	return 0;
}

static void free_tor(struct tor3 *tor)
{
	unsigned int x, f;

	for (x = 0; x < SPANS_PER_CARD; x++) {
		for (f = 0; f < (tor->tspans[x].cardtype == TYPE_E1 ? 31 : 24); f++) {
			if (tor->chans[x][f]) {
				kfree(tor->chans[x][f]);
			}
		}
		if (tor->chans[x])
			kfree(tor->chans[x]);
	}
	kfree(tor);
}

static int __devinit tor3_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int res,x,f;
	int ret = -ENODEV;
	struct tor3 *tor;
	unsigned long endjif;
	__iomem volatile unsigned long *gpdata_io, *lasdata_io;
	unsigned long gpdata,lasdata;

	res = pci_enable_device(pdev);
	if (res)
		return res;
	tor = kmalloc(sizeof(struct tor3), GFP_KERNEL);
	if (!tor)
		return -ENOMEM;
	memset(tor,0,sizeof(struct tor3));
	tor->lastmode = -1;
	spin_lock_init(&tor->lock);
	/* Load the resources */
	tor->pci = pdev;
	tor->irq = pdev->irq;
	if (tor->irq < 1) {
		printk(KERN_ERR "No IRQ allocated for device\n");
		goto err_out_free_tor;
	}
	tor->plx_region = pci_resource_start(pdev, 0);
	tor->plx_len = pci_resource_len(pdev, 0);
	tor->plx = ioremap(tor->plx_region, tor->plx_len);
	/* We don't use the I/O space, so we dont do anything with section 1 */
	tor->xilinx32_region = pci_resource_start(pdev, 2);
	tor->xilinx32_len = pci_resource_len(pdev, 2);
	tor->mem32 = ioremap(tor->xilinx32_region, tor->xilinx32_len);
	tor->xilinx8_region = pci_resource_start(pdev, 3);
	tor->xilinx8_len = pci_resource_len(pdev, 3);
	tor->mem8 = ioremap(tor->xilinx8_region, tor->xilinx8_len);
	/* Record what type */
	tor->type = (char *)ent->driver_data;
	/* Verify existence and accuracy of resources */
	if (!tor->plx_region || !tor->plx ||
	    (pci_resource_flags(pdev, 0) & IORESOURCE_IO)) {
		printk(KERN_ERR "Invalid PLX 9030 Base resource\n");
		goto err_out_free_tor;
	}
	if (!tor->xilinx32_region || !tor->mem32 ||
	    (pci_resource_flags(pdev, 2) & IORESOURCE_IO)) {
		printk(KERN_ERR "Invalid Xilinx 32 bit Base resource\n");
		goto err_out_free_tor;
	}
	if (!tor->xilinx8_region || !tor->mem8 ||
	    (pci_resource_flags(pdev, 3) & IORESOURCE_IO)) {
		printk(KERN_ERR "Invalid Xilinx 8 bit Base resource\n");
		goto err_out_free_tor;
	}
	/* Request regions */
	if (!request_mem_region(tor->plx_region, tor->plx_len, tor->type)) {
		printk(KERN_ERR "Unable to reserve PLX memory %08lx window at %08lx\n",
		       tor->plx_len, tor->plx_region);
		goto err_out_free_tor;
	}
	if (!request_mem_region(tor->xilinx32_region, tor->xilinx32_len, tor->type)) {
		printk(KERN_ERR "Unable to reserve Xilinx 32 bit memory %08lx window at %08lx\n",
		       tor->xilinx32_len, tor->xilinx32_region);
		goto err_out_release_plx_region;
	}
	if (!request_mem_region(tor->xilinx8_region, tor->xilinx8_len, tor->type)) {
		printk(KERN_ERR "Unable to reserve Xilinx memory %08lx window at %08lx\n",
		       tor->xilinx8_len, tor->xilinx8_region);
		goto err_out_release_plx_region;
	}
	pci_set_drvdata(pdev, tor);
	printk(KERN_INFO "Detected %s at 0x%lx/0x%lx irq %d\n", tor->type, 
		tor->xilinx32_region, tor->xilinx8_region,tor->irq);

	for (x = 0; x < MAX_TOR_CARDS; x++) {
		if (!cards[x]) break;
	}
	if (x >= MAX_TOR_CARDS) {
		printk(KERN_ERR "No cards[] slot available!!\n");
		goto err_out_release_all;
	}
	tor->num = x;
	cards[x] = tor;

	/* start programming mode */
	gpdata_io = (__iomem unsigned long *) &tor->plx[GPIOC];
	gpdata = le32_to_cpu(*gpdata_io);

	gpdata |= GPIO_WRITE; /* make sure WRITE is not asserted */
	*gpdata_io = cpu_to_le32(gpdata);

	gpdata &= ~GPIO_PROGRAM;  /* activate the PROGRAM signal */
	*gpdata_io = cpu_to_le32(gpdata);

	/* wait for INIT and DONE to go low */
	endjif = jiffies + 10;
	while (le32_to_cpu(*gpdata_io) & (GPIO_INIT | GPIO_DONE) && (jiffies <= endjif));

	if (endjif < jiffies) {
		printk(KERN_ERR "Timeout waiting for INIT and DONE to go low\n");
		goto err_out_release_all;
	}
	if (debug) printk(KERN_ERR "fwload: Init and done gone to low\n");
	gpdata |= GPIO_PROGRAM;
	*gpdata_io = cpu_to_le32(gpdata);  /* de-activate the PROGRAM signal */
	/* wait for INIT to go high (clearing done */
	endjif = jiffies + 10;
	while (!(le32_to_cpu(*gpdata_io) & GPIO_INIT) && (jiffies <= endjif));
	if (endjif < jiffies) {
		printk(KERN_ERR "Timeout waiting for INIT to go high\n");
		goto err_out_release_all;
	}

	if (debug) printk(KERN_ERR "fwload: Init went high (clearing done)\nNow loading...\n");
	/* assert WRITE signal */
	gpdata &= ~GPIO_WRITE;
	*gpdata_io = cpu_to_le32(gpdata);
	for (x = 0; x < sizeof(tor2fw); x++) {
		/* write the byte */
		*tor->mem8 = tor2fw[x];
		/* if DONE signal, we're done, exit */
		if (le32_to_cpu(*gpdata_io) & GPIO_DONE) break;
		/* if INIT drops, we're screwed, exit */
		if (!(le32_to_cpu(*gpdata_io) & GPIO_INIT)) break;
	}
	if (debug) printk(KERN_DEBUG "fwload: Transferred %d bytes into chip\n",x);
	/* Wait for FIFO to clear */
	endjif = jiffies + 2;
	while (jiffies < endjif); /* wait */
	/* de-assert write signal */
	gpdata |= GPIO_WRITE;
	*gpdata_io = cpu_to_le32(gpdata);
	if (debug) printk(KERN_DEBUG "fwload: Loading done!\n");	

	/* Wait for FIFO to clear */
	endjif = jiffies + 2;
	while (jiffies < endjif); /* wait */
	if (!(le32_to_cpu(*gpdata_io) & GPIO_INIT)) {
		printk(KERN_ERR "Drove Init low!! CRC Error!!!\n");
		goto err_out_release_all;
	}
	if (!(le32_to_cpu(*gpdata_io) & GPIO_DONE)) {
		printk(KERN_ERR "Did not get DONE signal. Short file maybe??\n");
		goto err_out_release_all;
	}
	printk(KERN_INFO "Xilinx Chip successfully loaded, configured and started!!\n");

	tor->mem8[SYNCREG] = 0;
	tor->mem8[CTLREG] = 0;
	tor->mem8[CTLREG1] = 0;
	tor->mem8[LEDREG] = 0;

	/* set the LA2BRD register so that we enable block transfer, read
	   pre-fetch, and set to maximum read pre-fetch size */
	lasdata_io = (__iomem unsigned long *) &tor->plx[LAS2BRD];
	lasdata = *lasdata_io;
	lasdata |= 0x39;
	*lasdata_io = lasdata;

	/* set the LA3BRD register so that we enable block transfer */
	lasdata_io = (__iomem unsigned long *) &tor->plx[LAS3BRD];
	lasdata = *lasdata_io;
	lasdata |= 1;
	*lasdata_io = lasdata;

	/* check part revision data */
	x = t1in(tor,1,0xf) & 15;
#ifdef	NONREVA
	if (x > 3) {
		tor->mem8[CTLREG1] = NONREVA;
	}
#endif
	for (x = 0; x < 256; x++) tor->mem32[x] = 0x7f7f7f7f;

	if (request_irq(tor->irq, tor3_intr, DAHDI_IRQ_SHARED_DISABLED, "tor3", tor)) {
		printk(KERN_ERR "Unable to request tormenta IRQ %d\n", tor->irq);
		goto err_out_release_all;
	}

	tor->lastmode = -1;
	if (strstr(tor->type, "TE400P")) {
		int i;
		for (i=0; i<SPANS_PER_CARD; i++) {
			if (mode & (1<<i)) {
				tor->tspans[i].cardtype = TYPE_T1;
			} else {
				tor->tspans[i].cardtype = TYPE_E1;
			}
		}
		tor->datxlt = datxlt_e1;
	} else {
		printk(KERN_ERR "Don't know that card\n");
		free_irq(tor->irq, tor);
		goto err_out_release_all;
	}

	for (x = 0; x < SPANS_PER_CARD; x++) {
		int num_chans = tor->tspans[x].cardtype == TYPE_E1 ? 31 : 24;
		
		if (!(tor->chans[x] = kmalloc(num_chans * sizeof(*tor->chans[x]), GFP_KERNEL))) {
			printk(KERN_ERR "tor3: Not enough memory for chans[%d]\n", x);
			ret = -ENOMEM;
			goto err_out_release_all;
		}
		for (f = 0; f < (num_chans); f++) {
			if (!(tor->chans[x][f] = kmalloc(sizeof(*tor->chans[x][f]), GFP_KERNEL))) {
				printk(KERN_ERR "tor3: Not enough memory for chans[%d][%d]\n", 
						x, f);
				ret = -ENOMEM;
				goto err_out_release_all;
			}
			memset(tor->chans[x][f], 0, sizeof(*tor->chans[x][f]));
		}
	}

	init_spans(tor); 

	tor->order = tor->mem8[SWREG];
	printk(KERN_INFO "Detected Card number: %d\n", tor->order);

	/* Launch cards as appropriate */
	x = 0;
	for (;;) {
		/* Find a card to activate */
		f = 0;
		for (x=0;cards[x];x++) {
			if (cards[x]->order <= highestorder) {
				tor3_launch(cards[x]);
				if (cards[x]->order == highestorder)
					f = 1;
			}
		}
		/* If we found at least one, increment the highest order and search again, otherwise stop */
		if (f) 
			highestorder++;
		else
			break;
	}

	return 0;

err_out_release_all:
	release_mem_region(tor->xilinx32_region, tor->xilinx32_len);
	release_mem_region(tor->xilinx8_region, tor->xilinx8_len);
err_out_release_plx_region:
	release_mem_region(tor->plx_region, tor->plx_len);
err_out_free_tor:
	if (tor->plx) iounmap(tor->plx);
	if (tor->mem8) iounmap(tor->mem8);
	if (tor->mem32) iounmap(tor->mem32);
	if (tor) {
		free_tor(tor);
	}
	return -ENODEV;
}

static struct pci_driver tor3_driver;

static void __devexit tor3_remove(struct pci_dev *pdev)
{
	struct tor3 *tor;
	int i;

	tor = pci_get_drvdata(pdev);
	if (!tor)
		BUG();
	tor->mem8[SYNCREG] = 0;
	tor->mem8[CTLREG] = 0;
	tor->mem8[LEDREG] = 0;
	tor->plx[INTCSR] = cpu_to_le16(0);
	free_irq(tor->irq, tor);
	for (i = 0; i < SPANS_PER_CARD; ++i) {
		if (test_bit(DAHDI_FLAGBIT_REGISTERED, &tor->spans[i].flags))
			dahdi_unregister(&tor->spans[i]);
	}
	release_mem_region(tor->plx_region, tor->plx_len);
	release_mem_region(tor->xilinx32_region, tor->xilinx32_len);
	release_mem_region(tor->xilinx8_region, tor->xilinx8_len);
	if (tor->plx) iounmap(tor->plx);
	if (tor->mem8) iounmap(tor->mem8);
	if (tor->mem32) iounmap(tor->mem32);

	cards[tor->num] = NULL;
	pci_set_drvdata(pdev, NULL);
	free_tor(tor);
}

static struct pci_driver tor3_driver = {
	.name = "tormenta2",
	.probe = tor3_probe,
	.remove = __devexit_p(tor3_remove),
	.id_table = tor3_pci_ids,
};

static int __init tor3_init(void) {
	int res;
	res = dahdi_pci_module(&tor3_driver);
	printk(KERN_INFO "Registered Tormenta2 PCI\n");
	return res;
}

static void __exit tor3_cleanup(void) {
	pci_unregister_driver(&tor3_driver);
	printk(KERN_INFO "Unregistered Tormenta2\n");
}

static void set_clear_t1(struct tor3 *tor, int s)
{
	int i,j;
	unsigned short val=0;
	for (i = 0; i < 24; i++) {
		j = (i/8);
		if (!(tor->spans[s].chans[i]->flags & DAHDI_FLAG_CLEAR))
			val |= 1 << (i % 8);
		
		if ((i % 8)==7) {
			t1out(tor,1 + s, 0x8 + j, val);
			val = 0;
		}
	}
}

static void set_clear_e1(struct tor3 *tor, int s)
{
	int i,j;
	unsigned short val;
	if (tor->spans[s].lineconfig & DAHDI_CONFIG_CCS)
		return;
	val = 1;
	for (i = 1; i < 16; i++) {
		j = (i/8);
		if (!(tor->spans[s].chans[i]->flags & DAHDI_FLAG_CLEAR))
			val |= 1 << (i % 8);
		if ((i % 8)==7) {
			t1out(tor,1 + s, 0x8 + j, val);
			val = 0;
		}
	}
	val = 1;
	for (i = 17; i < 32; i++) {
		j = (i/8);
		if (!(tor->spans[s].chans[i]->flags & DAHDI_FLAG_CLEAR))
			val |= 1 << (i % 8);
		if ((i % 8)==7) {
			t1out(tor,1 + s, 0x8 + j, val);
			val = 0;
		}
	}
}
 
static int tor3_rbsbits(struct dahdi_chan *chan, int bits)
{
	u_char m,c;
	int k,n,b;
	int all, byte;
	int location = 0;
	struct tor3_chan *p = chan->pvt;
	unsigned long flags;

 	n = chan->chanpos - 1;
	all = 0;
	if (p->tor->tspans[p->span].cardtype == TYPE_E1) {
		if (chan->chanpos == 16) return 0;
		if (chan->chanpos > 16) {
			n--;
		}
		location++;
	}
	location += 0x50 + n/2;
	k = p->span;
	b = (n / 8); /* get byte number */
	m = 1 << (n & 7); /* get mask */
	c = p->tor->txsigs[k][b];
	c &= ~m;  /* clear mask bit */
	/* set mask bit, if bit is to be set */
	if (bits & DAHDI_ABIT) c |= m;
	if (c & m)
		all |= (1<<3);
	p->tor->txsigs[k][b] = c;
	b += 3; /* now points to b bit stuff */
	/* get current signalling values */
	c = p->tor->txsigs[k][b];
	c &= ~m;  /* clear mask bit */
	/* set mask bit, if bit is to be set */
	if (bits & DAHDI_BBIT) c |= m;
	/* save new signalling values */
	if (c & m)
		all |= (1<<2);
	/* save new signalling values */
	p->tor->txsigs[k][b] = c;
	b += 3; /* now points to c bit stuff */
	/* get current signalling values */
	c = p->tor->txsigs[k][b];
	c &= ~m;  /* clear mask bit */
	/* set mask bit, if bit is to be set */
	if (bits & DAHDI_CBIT) c |= m;
	if (c & m)
		all |= (1<<1);

	/* save new signalling values */
	p->tor->txsigs[k][b] = c;
	b += 3; /* now points to d bit stuff */
	/* get current signalling values */
	c = p->tor->txsigs[k][b];
	c &= ~m;  /* clear mask bit */
	/* set mask bit, if bit is to be set */
	if (bits & DAHDI_DBIT) c |= m;
	if (c & m)
		all |= (1<<0);

	/* save new signalling values */
	p->tor->txsigs[k][b] = c;
	/* output them into the chip */
	spin_lock_irqsave(&p->tor->lock, flags);	
	byte = t1in(p->tor,k + 1,location);
	byte &= 0xf0 >> ((n%2)*4);
	byte |= all << ((n%2)*4);
	t1out(p->tor,k + 1,location,byte);
	spin_unlock_irqrestore(&p->tor->lock, flags);
	return 0;
}

static int tor3_shutdown(struct dahdi_span *span)
{
	int i;
	int tspan;
	int wasrunning;
	unsigned long flags;
	struct tor3_span *p = span->pvt;

	tspan = p->span + 1;
	if (tspan < 0) {
		printk(KERN_DEBUG "Tor3: Span '%d' isn't us?\n", span->spanno);
		return -1;
	}

	spin_lock_irqsave(&p->tor->lock, flags);
	wasrunning = span->flags & DAHDI_FLAG_RUNNING;

	span->flags &= ~DAHDI_FLAG_RUNNING;

	if (p->tor->tspans[p->span].cardtype == TYPE_E1) {
		for (i = 0; i < 192; i++)
			t1out(p->tor,tspan, i, 0);
	} else {
		for (i = 0; i < 160; i++)
			t1out(p->tor,tspan, i, 0);
	}
	if (wasrunning)
		p->tor->spansstarted--;
	spin_unlock_irqrestore(&p->tor->lock, flags);	
	if (!(p->tor->spans[0].flags & DAHDI_FLAG_RUNNING) &&
	    !(p->tor->spans[1].flags & DAHDI_FLAG_RUNNING) &&
	    !(p->tor->spans[2].flags & DAHDI_FLAG_RUNNING) &&
	    !(p->tor->spans[3].flags & DAHDI_FLAG_RUNNING))
		/* No longer in use, disable interrupts */
		p->tor->mem8[CTLREG] = 0;
	if (debug)
		printk(KERN_DEBUG"Span %d (%s) shutdown\n", span->spanno, span->name);
	return 0;
}


static int tor3_startup(struct dahdi_span *span)
{
	unsigned long endjif;
	int i, x;
	int tspan;
	unsigned long flags;
	char *coding = "";
	char *framing = "";
	char *crcing = "";
	int alreadyrunning;
	struct tor3_span *p = span->pvt;
	struct tor3 *tor = p->tor;

	tspan = p->span + 1;
	if (tspan < 0) {
		printk(KERN_DEBUG "Tor3: Span '%d' isn't us?\n", span->spanno);
		return -1;
	}

	spin_lock_irqsave(&p->tor->lock, flags);

	alreadyrunning = span->flags & DAHDI_FLAG_RUNNING;

	/* initialize the start value for the entire chunk of last ec buffer */
	for (i = 0; i < span->channels; i++) {
		memset(p->tor->ec_chunk1[p->span][i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
		memset(p->tor->ec_chunk2[p->span][i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
	}
	/* Force re-evaluation of the timing source */
	if (timingcable)
		p->tor->syncsrc = -1;

	if (p->tor->tspans[p->span].cardtype == TYPE_E1) { /* if this is an E1 card */
		int lc = span->lineconfig;
		if (!alreadyrunning) {
			p->tor->mem8[SYNCREG] = SYNCSELF;
			p->tor->mem8[CTLREG] = E1DIV;
			p->tor->mem8[LEDREG] = 0;
			/* reset */
			for (i = 0; i < 192; i++) 
				t1out(tor,tspan, i, 0);

			/* e1 mode */
			t1out(tor,tspan,0x0,(1<<1) /* e1 */);

			/* Set up for Interleaved Serial Bus operation in byte mode */
			/* Set up all the spans every time, so we are sure they are
			   in a consistent state. If we don't, a card without all
			   its spans configured misbehaves in strange ways. */
			/* IBO */
			for (x=0; x<4; x++) {
				t1out(tor,1+x,0xC5,(1<<3)|(1<<5)|x);
			}
			t1out(tor,tspan,0x70,(1<<2)|(1<<7)); /* TCSS1 + MCLKS */
			t1out(tor,tspan,0x1,(1<<4)|(1<<1)); /* RSIO + 1 is from O12.0 */
			t1out(tor,tspan,0x2,(1<<0)|(1<<1));
			t1out(tor,tspan,0x4f,(1<<0)|(1<<4));
			t1out(tor,tspan,0x35,(1<<4)|(1<<0)|(1<<2));	/* TSiS, TCRC4 (from O14.4), THDB3(from O14.6) */
			t1out(tor,tspan,0x36,(1<<2)); /* AEBE 36.2 */
			t1out(tor,tspan,0x33,(1<<3)|(1<<5)|(1<<6)); /* RCR4, RHDB3, RSM */
			t1out(tor,tspan,0x34,(1<<0)); /* RCL */
			t1out(tor,tspan,0x40,(1<<1)|(1<<2)); /* RCCS, TCCS */
			t1out(tor,tspan,0xd0,(1<<0)|(1<<1)|(1<<3)|(1<<4)); /* TAFR */
			t1out(tor,tspan,0xd1,(1<<0)|(1<<1)|(1<<2)|(1<<3)|(1<<4)|(1<<6)); /* TNAFR */
			t1out(tor,tspan,0x79,(1<<4)|(1<<7)|(1<<3)); /* trying the JACLK */
			t1out(tor,tspan,0x78,(1<<0)|(1<<5)|(1<<3)); /* LB0 is from O18.5 */
			/* MPS1 + RT1 + TT1 */
			t1out(tor,tspan,0x7b,(1<<0)|(1<<1)|(1<<2)|(1<<3)|(1<<5)); /* not sure about internal termination */
		}

		t1out(tor,tspan,0x33,(lc & DAHDI_CONFIG_CRC4 ? (1<<3) : 0) | (lc & DAHDI_CONFIG_HDB3 ? (1<<5) : 0) | (lc & DAHDI_CONFIG_CCS ? (1<<6) : 0));
		t1out(tor,tspan,0x35,(lc & DAHDI_CONFIG_CRC4 ? (1<<0) : 0) | (lc & DAHDI_CONFIG_HDB3 ? (1<<2) : 0));
		t1out(tor,tspan,0x40,(lc & DAHDI_CONFIG_CCS ? (1<<1)|(1<<2) : 0));

		coding = lc & DAHDI_CONFIG_CCS ? "CCS" : "CAS";
		framing = lc & DAHDI_CONFIG_HDB3 ? "HDB3" : "AMI";
		crcing = lc & DAHDI_CONFIG_CRC4 ? "/CRC4" : "/NOCRC4";

		if (!alreadyrunning) {			
			span->flags |= DAHDI_FLAG_RUNNING;
			p->tor->spansstarted++;

			/* enable interrupts */
			p->tor->mem8[CTLREG] = INTENA | E1DIV;
		}

		set_clear_e1(p->tor, p->span);
		spin_unlock_irqrestore(&p->tor->lock, flags);

		if (debug) {
			if (alreadyrunning) 
				printk(KERN_INFO "Tor3: Reconfigured span %d (%s/%s%s) 120 Ohms\n", span->spanno, coding, framing, crcing);
			else
				printk(KERN_INFO "Tor3: Startup span %d (%s/%s%s) 120 Ohms\n", span->spanno, coding, framing, crcing);
		}
	} else { /* is a T1 card */
		int lc = span->lineconfig;
		if (!alreadyrunning) {
			DOSYNCREG(p->tor, SYNCSELF);
			p->tor->mem8[CTLREG] = p->tor->e1div;
			p->tor->mem8[LEDREG] = 0;

			/* reset */
			for (i = 0; i < 192; i++) 
				t1out(tor,tspan, i, 0);

			/* t1 mode */
			t1out(tor,tspan,0x0,0x0 /* t1 */);
		
			/* Set up for Interleaved Serial Bus operation in byte mode */
			/* Set up all the spans every time, so we are sure they are
				 in a consistent state. If we don't, a card without all
				 its spans configured misbehaves in strange ways. */
			/* IBO */
 
			for (x=0; x<4; x++) {
				t1out(tor,1+x,0xC5,(1<<3)|(1<<5)|x);
			}
			t1out(tor,tspan,0x70,(1<<2)|(1<<7)); /* TCSS1 + MCLKS */
			t1out(tor,tspan,0x1,(1<<4)|(1<<1)); /* RSIO + 1 is from O12.0 */
			t1out(tor,tspan,0x2,(1<<0)|(1<<1));
			t1out(tor,tspan,0x4f,(1<<0)|(1<<4));
			t1out(tor,tspan,0x3,(1<<3)); /* SYNCC */
			t1out(tor,tspan,0x4,(1<<5)|(1<<6)); /* ESF B8ZS */
			t1out(tor,tspan,0x5,(1<<4) | (japan ? (1<<7) : 0)); /* TSSE */
			/* transmit B8ZS */
			t1out(tor,tspan,0x6,(1<<7));
			/* transmit ESF */
			t1out(tor,tspan,0x7,(1<<2));
			t1out(tor,tspan,0x40,0x0); /* RCCS, TCCS */
			t1out(tor,tspan,0x79,(1<<3) | (1<<4)|(1<<6)); /* JACLK off for E1 - check prescaler */
			spin_unlock_irqrestore(&p->tor->lock, flags);
	
			endjif = jiffies + 10;
			while (jiffies < endjif); /* wait 100 ms */

			spin_lock_irqsave(&p->tor->lock, flags);
	
			t1out(tor,tspan,0x79,(1<<3) | (1<<4)); /* JACLK off for E1 - check prescaler */
			t1out(tor,tspan,0x78,(1<<0)|(1<<3)); /* just enable, DSX-1 0 to 133 ft but may be more later FIXME */
			/* MPS1 + RT1 + TT1 */
			t1out(tor,tspan,0x7b,(1<<1)|(1<<3)|(1<<5)); /* not sure about internal termination */
			t1out(tor,tspan,0x7a,0x0); /* transmit,receive synch &*/
		}

		t1out(tor,tspan,0x4,(lc & DAHDI_CONFIG_ESF ? (1<<6) : 0) | (lc & DAHDI_CONFIG_B8ZS ? (1<<5) : 0) | (japan ? (1<<1) : 0)); /* ESF B8ZS */

		coding = lc & DAHDI_CONFIG_ESF ? "ESF" : "SF";
		framing = lc & DAHDI_CONFIG_B8ZS ? "B8ZS" : "AMI";

		if (!alreadyrunning) {
			span->flags |= DAHDI_FLAG_RUNNING;
			p->tor->spansstarted++;
			/* enable interrupts */
			p->tor->mem8[CTLREG] = INTENA | p->tor->e1div;
		}

		set_clear_t1(p->tor, p->span);
		spin_unlock_irqrestore(&p->tor->lock, flags);

		if (debug) {
			if (alreadyrunning) 
				printk(KERN_INFO "Tor3: Reconfigured span %d (%s/%s) LBO: %s\n", span->spanno, coding, framing, dahdi_lboname(span->txlevel));
			else
				printk(KERN_INFO "Tor3: Startup span %d (%s/%s) LBO: %s\n", span->spanno, coding, framing, dahdi_lboname(span->txlevel));
		}
	}
	if (p->tor->syncs[0] == span->spanno) printk(KERN_INFO "SPAN %d: Primary Sync Source\n",span->spanno);
	if (p->tor->syncs[1] == span->spanno) printk(KERN_INFO "SPAN %d: Secondary Sync Source\n",span->spanno);
	if (p->tor->syncs[2] == span->spanno) printk(KERN_INFO "SPAN %d: Tertiary Sync Source\n",span->spanno);
	if (p->tor->syncs[3] == span->spanno) printk(KERN_INFO "SPAN %d: Quaternary Sync Source\n",span->spanno);
	return 0;
}

static int tor3_maint(struct dahdi_span *span, int cmd)
{
	return 0;
}

static inline void tor3_run(struct tor3 *tor)
{
	int x,y;
	for (x = 0; x < SPANS_PER_CARD; x++) {
		if (tor->spans[x].flags & DAHDI_FLAG_RUNNING) {
			/* since the Tormenta 2 PCI is double-buffered, you
			   need to delay the transmit data 2 entire chunks so
			   that the transmit will be in sync with the receive */
			for (y=0;y<tor->spans[x].channels;y++) {
				dahdi_ec_chunk(tor->spans[x].chans[y], 
				    tor->spans[x].chans[y]->readchunk, 
					tor->ec_chunk2[x][y]);
				memcpy(tor->ec_chunk2[x][y],tor->ec_chunk1[x][y],
					DAHDI_CHUNKSIZE);
				memcpy(tor->ec_chunk1[x][y],
					tor->spans[x].chans[y]->writechunk,
						DAHDI_CHUNKSIZE);
			}
			dahdi_receive(&tor->spans[x]);
		}
	}
	for (x = 0; x < SPANS_PER_CARD; x++) {
		if (tor->spans[x].flags & DAHDI_FLAG_RUNNING)
			dahdi_transmit(&tor->spans[x]);
	}
}

#ifdef ENABLE_TASKLETS
static void tor3_tasklet(unsigned long data)
{
	struct tor3 *tor = (struct tor3 *)data;
	tor->taskletrun++;
	if (tor->taskletpending) {
		tor->taskletexec++;
		tor3_run(tor);
	}
	tor->taskletpending = 0;
}
#endif

static int syncsrc = 0;
static int syncnum = 0 /* -1 */;
static int syncspan = 0;
static spinlock_t synclock = SPIN_LOCK_UNLOCKED;

static int tor3_findsync(struct tor3 *tor)
{
	int i;
	int x;
	unsigned long flags;
	int p;
	int nonzero;
	int newsyncsrc = 0;			/* DAHDI span number */
	int newsyncnum = 0;			/* tor3 card number */
	int newsyncspan = 0;		/* span on given tor3 card */
	spin_lock_irqsave(&synclock, flags);
#if 1
	if (!tor->num) {
		/* If we're the first card, go through all the motions, up to 8 levels
		   of sync source */
		p = 1;
		while (p < 8) {
			nonzero = 0;
			for (x=0;cards[x];x++) {
				for (i = 0; i < SPANS_PER_CARD; i++) {
					if (cards[x]->syncpos[i]) {
						nonzero = 1;
						if ((cards[x]->syncpos[i] == p) &&
						    !(cards[x]->spans[i].alarms & (DAHDI_ALARM_RED | DAHDI_ALARM_BLUE | DAHDI_ALARM_LOOPBACK)) &&
							(cards[x]->spans[i].flags & DAHDI_FLAG_RUNNING)) {
								/* This makes a good sync source */
								newsyncsrc = cards[x]->spans[i].spanno;
								newsyncnum = x;
								newsyncspan = i + 1;
								/* Jump out */
								goto found;
						}
					}
				}		
			}
			if (nonzero)
				p++;
			else 
				break;
		}
found:		
		if ((syncnum != newsyncnum) || (syncsrc != newsyncsrc) || (newsyncspan != syncspan)) {
			syncnum = newsyncnum;
			syncsrc = newsyncsrc;
			syncspan = newsyncspan;
			if (debug) printk(KERN_DEBUG "New syncnum: %d, syncsrc: %d, syncspan: %d\n", syncnum, syncsrc, syncspan);
		}
	}
#endif	
	/* update sync src info */
	if (tor->syncsrc != syncsrc) {
		tor->syncsrc = syncsrc;
		/* Update sync sources */
		for (i = 0; i < SPANS_PER_CARD; i++) {
			tor->spans[i].syncsrc = tor->syncsrc;
		}
		if (syncnum == tor->num) {
			/* actually set the sync register */
			DOSYNCREG(tor, syncspan);
			if (debug) printk(KERN_DEBUG "Card %d, using sync span %d, master\n", tor->num, syncspan);
			tor->master = MASTER;	
		} else {
			/* time from the timing cable */
			DOSYNCREG(tor, SYNCEXTERN);
			tor->master = 0;
			if (debug) printk(KERN_DEBUG "Card %d, using Timing Bus, NOT master\n", tor->num);	
		}
	}
	spin_unlock_irqrestore(&synclock, flags);
	return 0;
}

DAHDI_IRQ_HANDLER(tor3_intr)
{
	int n, i, j, k, newsyncsrc;
	unsigned long rxword,txword;

	unsigned char c, rxc;
	unsigned char abits, bbits, cbits, dbits;
	struct tor3 *tor = (struct tor3 *) dev_id;	

	  /* make sure its a real interrupt for us */
	if (!(tor->mem8[STATREG] & INTACTIVE)) { /* if not, just return */
		return IRQ_NONE;
	}

	/* set outbit, interrupt enable, and ack interrupt */
	tor->mem8[CTLREG] = OUTBIT | INTENA | INTACK | tor->master | tor->e1div;

#if	0
	if (!tor->passno)
		printk(KERN_DEBUG "Interrupt handler\n");
#endif

	/* do the transmit output */
	for (n = 0; n < 31; n++) {
		for (i = 0; i < DAHDI_CHUNKSIZE; i++) {
			/* span 1 */
			txword = tor->spans[0].chans[tor->tspans[0].revdatxlt[n] - 1]->writechunk[i] << 24;
			/* span 2 */
			txword |= tor->spans[1].chans[tor->tspans[1].revdatxlt[n] - 1]->writechunk[i] << 16;
			/* span 3 */
			txword |= tor->spans[2].chans[tor->tspans[2].revdatxlt[n] - 1]->writechunk[i] << 8;
			/* span 4 */
			txword |= tor->spans[3].chans[tor->tspans[3].revdatxlt[n] - 1]->writechunk[i];
			/* write to part */
#ifdef FIXTHISFOR64
			tor->mem32[tor->datxlt[n] + (32 * i)] = txword;
#else
			tor->mem32[tor->datxlt[n] + (32 * i)] = cpu_to_le32(txword);
#endif
		}
	}

	/* Do the receive input */
	for (n = 0; n < 31; n++) {
		for (i = 0; i < DAHDI_CHUNKSIZE; i++) {
			/* read from */
#ifdef FIXTHISFOR64
		        rxword = tor->mem32[tor->datxlt[n] + (32 * i)];
#else
			rxword = le32_to_cpu(tor->mem32[tor->datxlt[n] + (32 * i)]);
#endif
			/* span 1 */
			tor->spans[0].chans[tor->tspans[0].revdatxlt[n] - 1]->readchunk[i] = rxword >> 24;
			/* span 2 */
			tor->spans[1].chans[tor->tspans[1].revdatxlt[n] - 1]->readchunk[i] = (rxword & 0xff0000) >> 16;
			/* span 3 */
			tor->spans[2].chans[tor->tspans[2].revdatxlt[n] - 1]->readchunk[i] = (rxword & 0xff00) >> 8;
			/* span 4 */
			tor->spans[3].chans[tor->tspans[3].revdatxlt[n] - 1]->readchunk[i] = rxword & 0xff;
		}
	}

	i = tor->passno & 15;
	/* if a E1, do the signalling */
	if ((i < 12) && (tor->tspans[i/3].cardtype == TYPE_E1)) {
		int bit[5];
		int checkbit = 0;
		k = (i / 3);	/* get span */
		n = (i % 3);	/* get base */
		for (j=0; j<10; j++) {
			i = (n * 10) + j;
			/* skip the channel 16 */
			if (i >= 15) i++;
			if (!(tor->spans[k].chans[i]->sig & DAHDI_SIG_CLEAR))
				checkbit |= (1<<j);
		}
		for (j=0; j<5; j++) {
			if (checkbit & ((1<<(2*j)) || (1<<(2*j+1)))) {
				bit[j] = t1in(tor,k+1, 0x61 + j + (5*n));
			}
		}
		for (j=0; j<10; j++) {
			if (!(checkbit & (1<<j)))
				continue;
			/* Get channel number */
			i = (n * 10) + j;
			/* skip the channel 16 */
			if (i >= 15) i++;
			rxc = 0;
			abits = bit[j/2] & (1 << (4*((j%2)+1) - 1));
			bbits = bit[j/2] & (1 << (4*((j%2)+1) - 2));
			cbits = bit[j/2] & (1 << (4*((j%2)+1) - 3));
			dbits = bit[j/2] & (1 << (4*((j%2)+1) - 4));
			if (abits) rxc |= DAHDI_ABIT;
			if (bbits) rxc |= DAHDI_BBIT;
			if (cbits) rxc |= DAHDI_CBIT;
			if (dbits) rxc |= DAHDI_DBIT;
			if (tor->spans[k].chans[i]->rxsig != rxc) {
				/* Check for changes in received bits */
				dahdi_rbsbits(tor->spans[k].chans[i], rxc);
			}
		}
	}

	/* if a T1, do the signalling */
	if ((i < 12) && (tor->tspans[i/3].cardtype == TYPE_T1)) {
		int bit[4];
		int checkbit = 0;
		k = (i / 3);	/* get span */
		n = (i % 3);	/* get base */
		for (j=0; j<8; j++) {
			i = (n * 8) + j;
			if (!(tor->spans[k].chans[i]->sig & DAHDI_SIG_CLEAR)) {
				checkbit |= (1<<j);
			}
		}
		for (j=0; j<4; j++) {
			if (checkbit & ((1<<(2*j)) || (1<<(2*j+1))))
				bit[j] = t1in(tor,k+1, 0x60 + j + (4*n));
		}
		for (j=0; j<8; j++) {
			if (!(checkbit & (1<<j)))
				continue;
			/* Get channel number */
			i = (n * 8) + j;
			rxc = 0;
			abits = bit[j/2] & (1 << (4*((j%2)+1) - 1));
			bbits = bit[j/2] & (1 << (4*((j%2)+1) - 2));
			cbits = bit[j/2] & (1 << (4*((j%2)+1) - 3));
			dbits = bit[j/2] & (1 << (4*((j%2)+1) - 4));
			if (abits) rxc |= DAHDI_ABIT;
			if (bbits) rxc |= DAHDI_BBIT;
			if (cbits) rxc |= DAHDI_CBIT;
			if (dbits) rxc |= DAHDI_DBIT;
			if (tor->spans[k].chans[i]->rxsig != rxc) {
				/* Check for changes in received bits */
				dahdi_rbsbits(tor->spans[k].chans[i], rxc);
			}
		}
	}

	for (i = 0; i < SPANS_PER_CARD; i++) { /* Go thru all the spans */
		/* if alarm timer, and it's timed out */
		if (tor->alarmtimer[i]) {
			if (!--tor->alarmtimer[i]) {
				/* clear recover status */
				tor->spans[i].alarms &= ~DAHDI_ALARM_RECOVER;
				if (tor->tspans[i].cardtype == TYPE_T1)
					t1out(tor,i + 1,0x5,(1<<4) | (japan ? (1<<7) : 0)); /* TSSE, YEL */
				dahdi_alarm_notify(&tor->spans[i]);  /* let them know */
			}
		}
	}

	i = tor->passno & 15;
	if ((i >= 10) && (i <= 13) && !(tor->passno & 0x30)) {
		j = 0;  /* clear this alarm status */
		i -= 10;
		t1out(tor, i+1,0x18, 0xff);
		c = t1in(tor,i+1, 0x18);  /* get the status */
		if (c & (0x1|0x4)) /* if red alarm */
			j |= DAHDI_ALARM_RED;
		if (c & 0x2) /* if blue alarm */
			j |= DAHDI_ALARM_BLUE;

		/* only consider previous carrier alarm state */
		tor->spans[i].alarms &= (DAHDI_ALARM_RED | DAHDI_ALARM_BLUE | DAHDI_ALARM_NOTOPEN);
		n = 1; /* set to 1 so will not be in yellow alarm if we dont
							care about open channels */
		/* if to have yellow alarm if nothing open */
		if (tor->spans[i].lineconfig & DAHDI_CONFIG_NOTOPEN) {
			/* go thru all chans, and count # open */
			for (n = 0,k = 0; k < tor->spans[i].channels; k++) {
				if (((tor->chans[i][k])->flags & DAHDI_FLAG_OPEN) || ((tor->chans[i][k])->flags & DAHDI_FLAG_NETDEV))
					n++;
			}
			/* if none open, set alarm condition */
			if (!n) j |= DAHDI_ALARM_NOTOPEN; 
		}
		/* if no more alarms, and we had some */
		if ((!j) && tor->spans[i].alarms)
			tor->alarmtimer[i] = DAHDI_ALARMSETTLE_TIME;

		if (tor->alarmtimer[i])
			j |= DAHDI_ALARM_RECOVER;
		/* if going into alarm state, set yellow alarm */
		if ((j) && (!tor->spans[i].alarms)) {
			if (tor->tspans[i].cardtype == TYPE_T1)
				t1out(tor,i + 1,0x5,(1<<4) | (1<<0) | (japan ? (1<<7) : 0)); /* TSSE, YEL */
		}
#if 0
		if (tor->tspans[i].cardtype == TYPE_T1) {
			if (c & (1<<3)) /* if yellow alarm */
				j |= DAHDI_ALARM_YELLOW;
			if (c & (1<<7))
				j &= ~DAHDI_ALARM_YELLOW;
		}
#endif
		if (tor->spans[i].maintstat || tor->spans[i].mainttimer) j |= DAHDI_ALARM_LOOPBACK;
		tor->spans[i].alarms = j;
		c = (LEDRED | LEDGREEN) << (2 * i);
		tor->leds &= ~c;  /* mask out bits for this span */
		/* light LED's if span configured and running */
		if (tor->spans[i].flags & DAHDI_FLAG_RUNNING) {
			if (j & DAHDI_ALARM_RED) tor->leds |= LEDRED << (2 * i);
			else if (j & DAHDI_ALARM_YELLOW) tor->leds |= (LEDRED | LEDGREEN) << (2 * i);
			else tor->leds |= LEDGREEN << (2 * i);
		}
		tor->mem8[LEDREG] = tor->leds;
		dahdi_alarm_notify(&tor->spans[i]);
	}
	if (!(tor->passno % 1000)) { /* even second boundary */
		/* do all spans */
		for (i = 1; i <= SPANS_PER_CARD; i++) {
			/* add this second's BPV count to total one */
			tor->spans[i - 1].bpvcount += t1in(tor,i,0x43) + (t1in(tor,i,0x42) << 8);
			if (tor->tspans[i].cardtype == TYPE_E1) {
				if (tor->spans[i - 1].lineconfig & DAHDI_CONFIG_CRC4) {
					tor->spans[i - 1].crc4count += t1in(tor,i,0x45) + (t1in(tor,i,0x44) << 8);
					tor->spans[i - 1].ebitcount += t1in(tor,i,0x49) + (t1in(tor,i,0x48) << 8);
				}
				tor->spans[i - 1].fascount += t1in(tor,i,0x47) + (t1in(tor,i,0x46) << 8);
			}
		}
	}
	if (!timingcable) {
		/* re-evaluate active sync src (no cable version) */
		tor->syncsrc = 0;
		newsyncsrc = 0;
		/* if primary sync specified, see if we can use it */
		if (tor->psyncs[0]) {
			/* if no alarms, use it */
			if (!(tor->spans[tor->psyncs[0] - 1].alarms & (DAHDI_ALARM_RED | DAHDI_ALARM_BLUE | DAHDI_ALARM_LOOPBACK))) {
				tor->syncsrc = tor->psyncs[0];
				newsyncsrc = tor->syncs[0];
			}
		}
		/* if any others specified, see if we can use them */
		for (i = 1; i < SPANS_PER_CARD; i++) {
			/* if we dont have one yet, and there is one specified at this level, see if we can use it */
			if ((!tor->syncsrc) && (tor->psyncs[i])) {
				/* if no alarms, use it */
				if (!(tor->spans[tor->psyncs[i] - 1].alarms & (DAHDI_ALARM_RED | DAHDI_ALARM_BLUE | DAHDI_ALARM_LOOPBACK))) {
					tor->syncsrc = tor->psyncs[i];
					newsyncsrc = tor->syncs[i];
				}
			}
		}
		/* update sync src info */
		for (i = 0; i < SPANS_PER_CARD; i++) tor->spans[i].syncsrc = newsyncsrc;

		/* actually set the sync register */
		DOSYNCREG(tor, tor->syncsrc);
	} else	/* Timing cable version */
		tor3_findsync(tor);

	tor->passno++;

#ifdef ENABLE_TASKLETS
	if (!tor->taskletpending) {
		tor->taskletpending = 1;
		tor->taskletsched++;
		tasklet_hi_schedule(&tor->tor3_tlet);
	} else {
		tor->txerrors++;
	}
#else
	tor3_run(tor);
#endif
	/* We are not the timing bus master */
	/* clear OUTBIT and enable interrupts */
	tor->mem8[CTLREG] = INTENA | tor->master | tor->e1div;
	return IRQ_RETVAL(1);
}


static int tor3_ioctl(struct dahdi_chan *chan, unsigned int cmd, unsigned long data)
{
	switch(cmd) {
	default:
		return -ENOTTY;
	}
	return 0;
}

MODULE_AUTHOR("Martin Pycko <mpycko@pbxhardware.com>");
MODULE_DESCRIPTION("Tormenta 2 PCI Quad T1 / E1 Zaptel Driver");
MODULE_LICENSE("GPL v2");

module_param(debug, int, 0600);
module_param(loopback, int, 0600);
module_param(mode, int, 0600);
module_param(timingcable, int, 0600);
module_param(japan, int, 0600);

MODULE_DEVICE_TABLE(pci, tor3_pci_ids);

module_init(tor3_init);
module_exit(tor3_cleanup);
